import java.util.concurrent.atomic.AtomicInteger;

class Cas extends Thread {
	public interface Inc {
		public void inc();
		public int get();
	}

	public static class MutexInc implements Inc{
		private int value;

		public MutexInc() {
			value = 0;
		}

		public void inc() {
			synchronized(this) {
				++value;
			}
		}

		public int get() {
			return value;
		}
	}

	public static class AtomicInc implements Inc {
		private AtomicInteger atomicInt;

		public AtomicInc() {
			atomicInt = new AtomicInteger(0);
		}

		public void inc() {
			atomicInt.getAndIncrement();
		}

		public int get() {
			return atomicInt.get();
		}
	}


	private Inc inc;
	private int limit;

	public Cas(int limit, Inc inc) {
		this.inc = inc;
		this.limit = limit;
	}

	public void run() {
		for (int i=0; i < limit; ++i) {
			inc.inc();
		}
	}

	private static long twoThreads(int limit, Inc inc) {
		Thread thread1 = new Cas(limit/2, inc);
		Thread thread2 = new Cas(limit/2, inc);

		final long t0 = System.currentTimeMillis();

		thread1.start(); 
		thread2.start();
		try { thread1.join();thread2.join();} catch (InterruptedException e) {}

		final long t = System.currentTimeMillis();

		return t - t0;
	}

	private static long singleThread(int limit, Inc inc) {
		Thread thread = new Cas(limit, inc);

		final long t0 = System.currentTimeMillis();
		thread.start();
		try { thread.join(); } catch (InterruptedException e) {}
		final long t = System.currentTimeMillis();

		return t - t0;
	}

	public static void main(String argv[]) {		
		final int limit = (argv.length == 1) ? Integer.parseInt(argv[0])
			                                 : 500000000;
		int counter= 0;

		long t0 = System.currentTimeMillis();
		for (int i=0; i < limit; ++i) {
			counter+=i;
		}

		long t = System.currentTimeMillis();
		System.out.println("single thread: " + (t - t0) + "ms" 
						   + "   result " + counter);

		System.out.println("single thread with lock: " 
						   + singleThread(limit, new MutexInc()) + "ms");

		System.out.println("two threads with lock: " 
						   + twoThreads(limit, new MutexInc()) + "ms");

		System.out.println("single thread with atomic inc: " 
						   + singleThread(limit, new AtomicInc()) + "ms");

		System.out.println("two threads with atmoic inc: " 
						   + twoThreads(limit, new AtomicInc()) + "ms");
	}
}