/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

compile 
g++ -O3 -o cas cas.cpp -lboost_chrono -lboost_program_options -lboost_thread 
*/

#include <boost/thread.hpp>
#include <boost/program_options.hpp>
#include <boost/chrono/chrono_io.hpp>
#include <boost/smart_ptr/detail/atomic_count.hpp>

#include <iostream>

namespace po = boost::program_options;
using namespace boost::chrono;

class MutexWorker {
  public:
    MutexWorker(boost::mutex& mutex, 
		   size_t from, size_t to,
		   size_t* counterPtr) 
	  : mutex_(mutex), 
		from_(from), to_(to),
		counterPtr_(counterPtr) { }
  
    void operator()() {
	    for (size_t i=from_; i < to_; ++i) {
            boost::mutex::scoped_lock lock(mutex_);
			++*counterPtr_;
		}
	}
  
  private:
    boost::mutex& mutex_;
    const size_t to_;
    const size_t from_;
    size_t* const counterPtr_;
};


class AtomicWorker {
  public:
    AtomicWorker(boost::detail::atomic_count& n,
				 size_t from, size_t to)
	  : n_(n), from_(from), to_(to) { }
  
    void operator()() {
	    for (size_t i=from_; i < to_; ++i) {
		    ++n_;
		}
	}
  
  private:
    boost::detail::atomic_count& n_;
    const size_t to_;
    const size_t from_;
};


int main(int argc, char* argv[]) {
    po::options_description desc("Allowed options");
	desc.add_options()
	    ("help", "produce help message")
	    ("increments", po::value<size_t>(), "number of atomic increments");

	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);    
	
	if (vm.count("help")) {
	    std::cout << desc << std::endl;
	    return 1;
	}

	size_t limit = 500000000;
	if (vm.count("increments")) {
	    limit = vm["increments"].as<size_t>();
	}

	std::cout << "limit is " << limit << std::endl;

	// single thread
	size_t counter = 0;
  	high_resolution_clock::time_point t0 = high_resolution_clock::now();

	for (size_t i=0; i < limit; ++i) {
	    counter+=i; // don't use ++ here, because gcc will optimise the loop
	}

  	high_resolution_clock::time_point t = high_resolution_clock::now();
	std::cout << "single thread: " << (t - t0) 
              << "  result " << counter << std::endl;

	// single thread with lock
	counter = 0;
	boost::mutex mutex;
	t0 = high_resolution_clock::now();

	boost::thread w(MutexWorker(mutex, 0, limit, &counter));
	w.join(); 

	t = high_resolution_clock::now();
	std::cout << "single thread with lock: " << (t - t0) << std::endl;

	// two threads with lock
	counter = 0;
	t0 = high_resolution_clock::now();

	boost::thread w1(MutexWorker(mutex, 0, limit/2, &counter));
	boost::thread w2(MutexWorker(mutex, limit/2, limit, &counter));

	w1.join(); 
	w2.join();

	t = high_resolution_clock::now();
	std::cout << "two threads with lock: " << (t - t0) << std::endl;

	// one thread with atomic
	boost::detail::atomic_count n(0);
	t0 = high_resolution_clock::now();
	
	for (size_t i=0; i < limit; ++i) {
	    ++n;
	}	
	t = high_resolution_clock::now();
	std::cout << "one thread atomic inc: " << (t - t0) << std::endl;


	// two threads with atomic
	boost::detail::atomic_count m(0);
	t0 = high_resolution_clock::now();

	w1 = boost::thread(AtomicWorker(m, 0, limit/2));
	w2 = boost::thread(AtomicWorker(m, limit/2, limit));

	w1.join(); 
	w2.join();

	t = high_resolution_clock::now();
	std::cout << "two thread atomic inc: " << (t - t0) << std::endl;
}
